// Highlevel handling of a session to a LysKOM server.
//
// $Id: session.cc,v 1.48 1996/08/26 10:00:23 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stddef.h>
#include <iostream.h>
#include <assert.h>

#include "session.h"

#include "login_q.h"
#include "logout_q.h"
#include "get_time_q.h"
#include "get_last_text_q.h"
#include "get_info_q.h"
#include "set_motd_of_lyskom_q.h"
#include "set_presentation_q.h"
#include "change_what_i_am_doing_q.h"
#include "lookup_name_q.h"
#include "lookup_z_name_q.h"
#include "re_lookup_person_q.h"
#include "set_last_read_q.h"
#include "set_unread_q.h"
#include "get_marks_q.h"
#include "who_am_i_q.h"
#include "who_is_on_ident_q.h"
#include "create_person_q.h"
#include "add_member_q.h"
#include "sub_member_q.h"
#include "send_message_q.h"
#include "set_client_version_q.h"
#include "membership-cache.h"
#include "text-mapping.h"
#include "time-utils.h"
#include "user_active_q.h"

// ================================================================ 
//                    Constructors and Destructors                  


Session::Session(Connection *connection,
		 const LyStr &client_name,
		 const LyStr &client_version) 
    : conn(connection),
      logged_in_person(0),
      session_number(0),
      all_confs(NULL),
      all_persons(NULL),
      all_maps(NULL),
      all_texts(NULL),
      my_user_area(NULL),
      all_memberships(NULL),
      previous_user_active_time(0),
      pending_user_active(NULL),
      ignored_questions()
{
    assert(connection != NULL);

    if (conn->get_server_info()->get_version() >= 10400)
    {
	question *q;
	q = new set_client_version_question(conn, client_name, client_version);
	ignored_questions.push_back(q);
    }

    m_async_dispatcher = new Async_dispatcher(connection);
}
 
// Session::Session(const Session &s)
//     : conn(s.conn), all_confs(s.all_confs), all_persons(s.all_persons), 
//       all_maps(s.all_maps), all_texts(s.all_texts),
//       logged_in_person(s.logged_in_person), session_number(s.session_number)
// {
// }


Session::~Session()
{
    if (my_user_area != NULL)
	delete my_user_area;
    for (long i = 0; i < ignored_questions.size(); ++i)
	delete ignored_questions[i];
    if (all_confs != NULL)
	delete all_confs;
    if (all_persons != NULL)
	delete all_persons;
    if (all_maps != NULL)
	delete all_maps;
    if (all_texts != NULL)
	delete all_texts;
    if (all_memberships != NULL)
	delete all_memberships;
    if (m_async_dispatcher != NULL)
	delete m_async_dispatcher;
}


// ================================================================ 
//

// const Session & 
// Session::operator = (const Session &s)
// {
//     conn    = s.conn;
//     all_confs   = s.all_confs;
//     all_persons = s.all_persons;
//     all_maps    = s.all_maps;
//     all_texts   = s.all_texts;
// 
//     logged_in_person = s.logged_in_person;
//     session_number   = s.session_number;
// 
//     return *this;
// }


// ostream &
// operator <<(ostream &os, const Session &s)
// {
//     cout << "A session object";
//     return os;
// }


//void
//Session::set_connection(Connection *connection)
//{
//    conn = connection;
//    //+++ Reset confs, persons, maps and texts here.
//}


// ================================================================


Conferences & 
Session::conferences()
{
    assert(conn != NULL);

    if (all_confs == NULL) {
	all_confs = new Conferences(conn, m_async_dispatcher);
    }

    return *all_confs;
}


Persons &
Session::persons()
{
    if (all_persons == NULL) {
	all_persons = new Persons(conferences());
    }

    return *all_persons;
}


Maps &
Session::maps()
{
    if (all_maps == NULL) {
	all_maps = new Maps(conferences(), m_async_dispatcher);
    }

    return *all_maps;
}


Texts &  
Session::texts()
{
    assert(conn != NULL);

    if (all_texts == NULL) {
	all_texts = new Texts(conn, m_async_dispatcher);
    }

    return *all_texts;

}


// ----------------------------------------------------------------

Status
Session::create_person(const LyStr &name,
		       const LyStr &password,
		       Pers_no *result,
		       bool *autologin)
{
    create_person_question q(conn, name, password);
    
    Status ret = q.receive();
    if (ret == st_ok)
    {
	if (result != NULL)
	    *result = q.result();
	// If you are not already logged in, an automatic login occurs.
	if (logged_in_person == 0)
	{
	    logged_in_person = q.result();
	    if (autologin != NULL)
		*autologin = true;
	}
	else
	    if (autologin != NULL)
		*autologin = false;
    }

    return ret;
}

Status
Session::login(Pers_no person, const LyStr &passwd, bool invisible)
{
    login_question in(conn, person, passwd, invisible);

    // +++ FIXME: Delete all caches!  Should this be a separate method?'
    if (my_user_area != NULL)
    {
	delete my_user_area;
	my_user_area = NULL;
    }

    Status ret = in.receive();
    if (ret == st_ok)
	logged_in_person = person;
    else
	logged_in_person = 0;

    return ret;
}

Status
Session::logout()
{
    if (logged_in_person != 0)
    {
	if (my_user_area != NULL)
	{
	    delete my_user_area;
	    my_user_area = NULL;
	}

	// FIXME: Delete all caches!
	logout_question out(conn);
	logged_in_person = 0;
	return out.receive();
    }
    else
	return st_ok;
}
    
Status
Session::add_member(Conf_no cno, 
		    Pers_no reader,
		    unsigned char priority, 
		    unsigned short placement)
{
    add_member_question q(conn, cno, reader, priority, placement);

    if (all_memberships == NULL)
	all_memberships = new Membership_cache(conn);

    if (all_memberships->present_or_pending(reader, cno))
	all_memberships->get(reader, cno).invalidate();

    return q.receive();
}

Kom_err
Session::sub_member(Conf_no cno, 
		    Pers_no member)
{
    sub_member_question q(conn, cno, member);

    if (all_memberships != NULL
	&& all_memberships->present_or_pending(member, cno))
	all_memberships->get(member, cno).set(Membership(KOM_NOT_MEMBER));

    if (q.receive() == st_ok)
	return KOM_NO_ERROR;
    else
	return q.error();
}

Status
Session::send_message(Pers_no recipient, 
		      const LyStr &message)
{
    // FIXME: Loses error code.
    send_message_question q(conn, recipient, message);
    return q.receive();
}

struct tm
Session::get_time()
{
    get_time_question   timeq(conn);
    Status              stat = timeq.receive();

    if (stat == st_ok)
	return timeq.result();
    else
	abort();		// +++ Error handling!!
}


Status
Session::change_what_i_am_doing(const LyStr &str)
{
    check_ignored();
    // FIXME: should the result maybe be ignored, so that no ping-pong
    // effect exists here?
    change_what_i_am_doing_question  doing(conn, str);

    return doing.receive();
}


Array<Who_info_ident>
Session::who_is_on()
{
    who_is_on_ident_question   who_q(conn);
    
    if (who_q.receive() == st_ok)
	return who_q.result();
    else
	abort();		// +++Error handling!
}

Server_info
Session::get_server_info()
{
    return *conn->get_server_info();
}


Kom_err
Session::set_presentation(Conf_no cno, Text_no tno)
{
    set_presentation_question  q(conn, cno, tno);

    if (q.receive() == st_ok)
    {
	// FIXME: Update the presentation field in the local cache.
	return KOM_NO_ERROR;
    }
    else
	return q.error();
}


Status
Session::set_motd_of_lyskom(Text_no motd)
{
    set_motd_of_lyskom_question  q(conn, motd);

    return q.receive();
}


Kom_res<Text_no>
Session::get_last_text(const struct tm &tme)
{
    get_last_text_question q(conn, tme);

    if (q.receive() == st_error)
	return q.error();
    else
	return q.result();
}

bool
Session::is_supervisor(Pers_no supervisor, Conf_no cno)
{
    if (supervisor == 0)
	return false;

    if (supervisor == cno)
	return true;

    Conference conf = conferences()[cno];
    if (!conf.exists())
	return false;

    if (conf.supervisor() == 0)
	return false;

    if (conf.supervisor() == supervisor)
	return true;

    if (is_member(supervisor, conf.supervisor()))
	return true;

    return false;
}

bool
Session::is_member(Pers_no person, Conf_no conf)
{
    if (all_memberships == NULL)
	all_memberships = new Membership_cache(conn);

    Membership m(all_memberships->get(person, conf));
    return m.errno() == KOM_NO_ERROR;
}

Status
Session::prefetch_is_member(Pers_no person, Conf_no conf)
{
    if (all_memberships == NULL)
	all_memberships = new Membership_cache(conn);

    return all_memberships->prefetch(person, conf);
}

Membership
Session::get_membership(Pers_no person, Conf_no conf)
{
    if (all_memberships == NULL)
	all_memberships = new Membership_cache(conn);

    Membership m(all_memberships->get(person, conf));
    return m;
}


Status
Session::prefetch_membership(Pers_no person, Conf_no conf)
{
    if (all_memberships == NULL)
	all_memberships = new Membership_cache(conn);

    return all_memberships->prefetch(person, conf);
}


bool
Session::prefetch_membership_pending(Pers_no person, Conf_no conf)
{
    if (all_memberships == NULL)
	all_memberships = new Membership_cache(conn);

    return all_memberships->prefetch_pending(person, conf);
}


bool
Session::membership_present(Pers_no person, Conf_no conf)
{
    if (all_memberships == NULL)
	all_memberships = new Membership_cache(conn);

    return all_memberships->present(person, conf);
}


Kom_err
Session::set_last_read(Conf_no cno,
		       Local_text_no last_text_read)
{
    if (is_true(server_supports_set_last_read)
	|| (!is_false(server_supports_set_last_read)
	    && conn->get_server_info()->get_version() >= 10800))
    {
	// Attempt to use this new call.
	set_last_read_question q(conn, cno, last_text_read);
	if (q.receive() != st_error || q.error() != KOM_NOT_IMPL)
	{
	    server_supports_set_last_read = true;
	    // The server apparently supported it.
	    if (q.receive() == st_ok)
		return KOM_NO_ERROR;
	    else
		return q.error();
	}
    }
    
    server_supports_set_last_read = false;
    // Fallback: use an older question that requries more preprocessing.
    Conference conf = conferences()[cno];
    if (conf.exists()) 
    {
	Local_text_no highest = conf.first_local_no() + conf.no_of_texts() - 1;
	int wanted = highest - last_text_read;
	if (wanted < 0)
	    wanted = 0;
	set_unread_question fallback(conn, cno, wanted);
	if (fallback.receive() == st_ok)
	    return KOM_NO_ERROR;
	else
	    return fallback.error();
    }
    else
    {
	return conf.errno();
    }
}

Array<Micro_conf>
Session::lookup_name(const LyStr &pattern)
{
    lookup_name_question   lookupq(conn, pattern);

    if (lookupq.receive() == st_ok) 
	return lookupq.result();
    else
	abort();		// +++Error handling!
}

void
Session::user_active()
{
    // Receive any pending answer to a previous user-active question.
    if (pending_user_active != NULL
	&& pending_user_active->status() != st_pending)
    {
	if (pending_user_active->status() == st_error
	    && (pending_user_active->error() == KOM_NOT_IMPL
		|| pending_user_active->error() == KOM_OBSOLETE))
	{
	    server_supports_user_active = false;
	}

	delete pending_user_active;
	pending_user_active = NULL;
    }

    if (is_true(server_supports_user_active)
	|| (!is_false(server_supports_user_active)
	    && conn->get_server_info()->get_version() >= 10900))
    {
	if (pending_user_active == NULL
	    && !is_false(server_supports_user_active))
	{
	    time_t now = time(NULL);

	    if (previous_user_active_time == 0 ||
		now - previous_user_active_time > 30)
	    {
		server_supports_user_active = true;
		previous_user_active_time = now;
		pending_user_active = new user_active_question(conn);
	    }
	}
    }
}

Array<Conf_z_info>
Session::lookup_z_name(const LyStr &pattern,
		       bool want_persons,
		       bool want_confs)
{
    lookup_z_name_question   lookupq(conn, pattern, want_persons, want_confs);

    if (lookupq.receive() == st_ok) 
	return lookupq.result();
    else
	abort();		// +++Error handling!
}


Array<Pers_no>
Session::regexp_lookup_persons(const LyStr &pattern)
{
    re_lookup_person_question   lookupq(conn, pattern);

    if (lookupq.receive() == st_ok) 
	return lookupq.result();
    else
	abort();		// +++Error handling!
}


Array<Mark>
Session::get_marks()
{
    get_marks_question  markq(conn);

    if (markq.receive() == st_ok) 
	return markq.result();
    else
	abort();		// +++Error handling!
}

Kom_res<Local_text_no>
Session::last_local_before(Conf_no cno, const struct tm &wanted_time)
{
    Conference conf = conferences()[cno];
    if (conf.errno() != KOM_NO_ERROR)
    {
	// return Kom_err<Local_text_no>(conf.errno());
	return conf.errno();
    }

    // FIXME: this could be rewritten more efficiently if Text_mapping
    // were a little smarter.  I could use a method that gives me the
    // next non-zero text, for instance.

    Text_mapping map = maps()[cno];

    Local_text_no lower = conf.first_local_no();
    Local_text_no higher = lower + conf.no_of_texts() - 1;
    Local_text_no diff = higher - lower;

    while (lower < higher)
    {
	Local_text_no middle = (lower + higher + 1)/2;
	Local_text_no attempt = middle;

	Text text = texts()[map[attempt]];
	while (!text.exists() && attempt < higher)
	    text = texts()[map[++attempt]];

	if (!text.exists())
	    higher = middle - 1; // No texts above middle exists.
	else
	{
	    struct tm entry_time = text.entry_time(cno, attempt);
	    if (struct_tm_newer(wanted_time, entry_time))
		lower = attempt;
	    else
		higher = middle - 1;
	}

	// Make sure the intervall was made smaller.
	assert(higher - lower < diff);
	diff = higher - lower;
    }

    while (lower >= conf.first_local_no())
    {
	Text found = texts()[map[lower]];
	if (found.exists() && struct_tm_newer(wanted_time,
					      found.entry_time(cno, lower)))
	    break;
	--lower;
    }

    return lower;
}

Kom_res<User_area*>
Session::get_user_area()
{
    if (my_user_area == NULL)
    {
	if (logged_in_person == 0)
	    return KOM_LOGIN;
	Texts &t = texts();
	my_user_area = new User_area(persons()[logged_in_person],
				     t.read_text_stat_cache(),
				     t.read_text_mass_cache());
    }

    return my_user_area;
}

				     
Pers_no
Session::my_login() const
{
    return logged_in_person;
}

Session_no
Session::my_session() const
{
    if (session_number == 0)
    {
	who_am_i_question wq(conn);
	if (wq.receive() == st_ok)
	    ((Session*)this)->session_number = wq.result();
	else
	    abort();		// FIXME: Error handling
    }

    return session_number;
}


Async_dispatcher &
Session::async_dispatcher()
{
    return *m_async_dispatcher;
}


Status
Session::handle_async(int this_many)
{
    check_ignored();
    return m_async_dispatcher->handle_async(this_many);
}

Connection *
Session::connection() const
{
    return conn;
}

void
Session::check_ignored()
{
    while (ignored_questions.size() > 0
	   && ignored_questions[0]->status() != st_pending)
    {
	delete ignored_questions[0];
	ignored_questions.pop_front();
    }
}
