// Handle information about conferences with unread texts.
//
// $Id: unread-confs.cc,v 1.43 1996/09/06 13:13:44 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <iostream.h>
#include <stdlib.h>

#include "unread-confs.h"
#include "connection.h"
#include "get_unread_confs_q.h"
#include "change_conference_q.h"
#include "async_text_created.h"

Unread_info::Unread_info()
    :conf(0), min_unread(-1), max_unread(-2)
{
}

Unread_info &
Unread_info::operator =(const Unread_info &u)
{
  conf = u.conf;
  min_unread = u.min_unread;
  max_unread = u.max_unread;
  return *this;
}

Unread_info_estimated::Unread_info_estimated()
    :conf(0), unread(-2)
{
}

Unread_info_estimated &
Unread_info_estimated::operator =(const Unread_info_estimated &u)
{
  conf = u.conf;
  unread = u.unread;
  return *this;
}

// ================================================================ 
//                    Constructors and Destructors                  


Unread_confs::Unread_confs(Session *sess, Pers_no the_reader)
    : session(sess),
      conferences(sess->conferences()), 
      persons(sess->persons()),
      maps(sess->maps()),
      texts(sess->texts()),
      unread_confs(), 
      read_confs(),
      reader(the_reader),
      current_read_order(NULL),
      ucq(NULL),
      cr_tag(sess->async_dispatcher().register_text_created_handler
	     (async_callback_method(this, &Unread_confs::new_text))),
      unread_confs_has_been_fetched(false)
{
}
 
// FIXME: Make Unread_confs into a reference counting class.
// This class owns the Read_orders pointed to by read_confs and unread_confs,
// so we cannot have two Unread_confs objects pointing to the same Read_orders.
//
// Unread_confs::Unread_confs(const Unread_confs &cl)
//     : session(cl.session), conferences(cl.conferences), persons(cl.persons),
//       maps(cl.maps), texts(cl.texts), unread_confs(cl.unread_confs), 
//       reader(cl.reader), read_confs(cl.read_confs),
//       current_read_order(cl.current_read_order), ucq(cl.ucq),
//       unread_confs_has_been_fetched(cl.unread_confs_has_been_fetched)
// {
// }


Unread_confs::~Unread_confs()
{
    Pix ix;

    for (ix = unread_confs.first(); ix != 0; unread_confs.next(ix))
	delete unread_confs(ix);
    for (ix = read_confs.first(); ix != 0; read_confs.next(ix))
	delete read_confs(ix);

    session->async_dispatcher().unregister_text_created_handler(cr_tag);

    delete ucq;
}


// ================================================================ 
//

//const Unread_confs & 
//Unread_confs::operator = (const Unread_confs &)
//{
//    cerr << "Operator= for class Unread_confs is not yet implemented\n";
//    return *this;
//}


// ================================================================
//

Status
Unread_confs::prefetch(Work_quota *work)
{
    int q = work->questions_left();
    // FIXME: there may be cases where there is nothing to prefetch.
    // This method should return st_ok in those cases.
    if (q == 0)
	return st_pending;

    if (!unread_confs_has_been_fetched)
    {
	ucq = new get_unread_confs_question(conferences.connection(), reader);
	unread_confs_has_been_fetched = true;
    }

    if (ucq != NULL && ucq->receive() == st_pending)
    {
	work->bump_question();
	return st_pending;
    }

    if (ucq != NULL)
    {
	if (ucq->receive() == st_error)
	   return st_error;

	assert(unread_confs.empty() && read_confs.empty());

	Array<Conf_no> new_unread_confs = ucq->result();
	delete ucq;
	ucq = NULL;

	for (int conf_ix = 0; conf_ix < new_unread_confs.size(); conf_ix++)
	{
	    add_conference(new_unread_confs[conf_ix], true);
	}
    }

    Pix ix = unread_confs.first();
    while (ix != 0 && !work->done())
    {
	work->change_conf();
	if (work->done())
	    break;

	Status st = unread_confs(ix)->prefetch(work);
	int nt = unread_confs(ix)->min_num_unread();

	if (st == st_ok && nt == 0)
	{     
	    // There is nothing to read in this conference. Remove it
	    // from unread_confs.
	    read_confs.append(unread_confs(ix));
	    unread_confs.del(ix);
	}
	else
	{
	    unread_confs.next(ix);
	}
    }

    return (q == work->questions_left()) ? st_ok : st_pending;
}

Conf_no
Unread_confs::goto_next_conf()
{
    if (goto_conf(query_next_conf()) != st_ok)
	return 0;
    return current_conf();
}


Conf_no
Unread_confs::query_next_conf()
{
    if (need_one_mapped() == st_error)
	return 0;

    if (unread_confs.empty())
	if (current_read_order == NULL)
	    return 0;
	else
	    return current_read_order->confno();

    // Unread_confs::prefetch should remove any empty Read_order.
    assert (unread_confs.front()->empty() == false);

    // Store the current Read_order in current if it contains unread texts.
    Read_order *current = NULL;
    if (current_read_order == unread_confs.front())
	current = unread_confs.remove_front(); 

    // Make sure that at least one text exists in the new current
    // working conference.
    need_one_mapped();

    Conf_no result = 0;
    if (!unread_confs.empty())
	result = unread_confs.front()->confno();

    if (current != NULL)
    {
	unread_confs.prepend(current);
	if (result == 0)
	    result = current->confno();
    }

    return result;
}

Status
Unread_confs::goto_conf(Conf_no c)
{
    if (current_read_order == NULL && c == 0)
	return st_ok;

    if (current_read_order != NULL && current_read_order->confno() == c)
	// Return immediately if we are already standing in the conference.
	return st_ok;

    Read_order *ro = NULL;

    if (c != 0)
    {
	Pix ix;

	if (need_unread_confs() != st_ok)
	    return st_error;

	for (ix = unread_confs.first(); ix != 0; unread_confs.next(ix))
	    if (unread_confs(ix)->confno() == c)
	    {
		ro = unread_confs(ix);
		unread_confs.del(ix);
		break;
	    }

	if (ro == NULL)
	    for (ix = read_confs.first(); ix != 0; read_confs.next(ix))
		if (read_confs(ix)->confno() == c)
		{
		    ro = read_confs(ix);
		    read_confs.del(ix);
		    break;
		}

	if (ro == NULL)
	    ro = new Read_order(session, c, reader);
    }

    change_conference_question q(conferences.connection(), c);
    // +++ We could relatively safely ignore this question, except
    // when we had to create a new Read_order.
    if (q.receive() == st_ok)
    {
	if (ro != NULL)
	    unread_confs.prepend(ro);
	current_read_order = ro;
    }
    else
    {
	// The user was not a member in this conference (or something
	// else happened).  ro may be NULL here if we are going to
	// conference 0.
	assert(current_read_order != ro);
	delete ro;
    }
    return q.receive();
}

Prompt
Unread_confs::next_prompt()
{
    Prompt   prompt;

    if (need_one_mapped() == st_error)
	return PROMPT_ERROR;

    if (unread_confs.empty())
	return NO_MORE_TEXT;

    if (current_read_order != unread_confs.front())
	return NEXT_CONF;

    prompt = unread_confs.front()->next_prompt();
    assert(prompt != NO_MORE_TEXT && prompt != NEXT_CONF);
    return prompt;
}


Text_no
Unread_confs::next_text(bool remove)
{
    assert(current_read_order != NULL);

    if (need_one_mapped() == st_error)
	return 0;

    assert(current_read_order != NULL);

    if (current_read_order->empty())
	return 0;

    Text_no  next = current_read_order->next_text(remove);
    if (next != 0 && remove)
	remove_text(next);

    return next;
}


void
Unread_confs::enqueue_children(Text_no tno)
{
    assert(current_read_order != NULL);
    current_read_order->enqueue_children(tno);
}


void
Unread_confs::remove_text(Text_no tno)
{
    if (need_unread_confs() != st_ok)
	return;

#if 1				// FIXME: Create test-case for this.
    // Silently ignore any attempts to remove a nonexisting text.
    if (!texts[tno].exists())
      return;
#endif

    // Loop through the recipients of the text.
    for (int i = 0; i < texts[tno].num_recipients(); ++i) {
	Conf_no  cno = texts[tno].recipient(i)->recipient();
	Local_text_no lno = texts[tno].recipient(i)->local_no();

        // Loop through all unread_confs and if we find the recipient,
        // then remove the text from it.
	Pix ix;
        for (ix = unread_confs.first(); ix != 0; unread_confs.next(ix)) {
	    if (unread_confs(ix)->confno() == cno)
		unread_confs(ix)->remove_text(tno, lno);
	}
    }
}


Conf_no
Unread_confs::current_conf()
{
    if (current_read_order == NULL)
	return 0;
    else
	return current_read_order->confno();
}

Array<Unread_info>
Unread_confs::confs_with_unread()
{
    if (need_unread_confs() != st_ok)
	return 0;

    Array<Unread_info>   arr(unread_confs.length());
    int                  index = 0;

    for (Pix ix = unread_confs.first(); ix != 0; unread_confs.next(ix)) 
    {
	arr[index].conf = unread_confs(ix)->confno();
	arr[index].min_unread = unread_confs(ix)->min_num_unread();
	arr[index].max_unread = unread_confs(ix)->max_num_unread();
	index++;
    }

    return arr;
}

Array<Unread_info_estimated>
Unread_confs::confs_with_estimated_unread(int max_limit)
{
    if (need_unread_confs() != st_ok)
	return 0;

    int m = unread_confs.length();
    if (max_limit != 0 && m > max_limit)
	m = max_limit;
    Array<Unread_info_estimated>   arr(m);
    int index = 0;

    // FIXME: it is probably possible to boost performance with
    // creative prefetching here.

    for (Pix ix = unread_confs.first(); ix != 0 && index < m;
	 unread_confs.next(ix))
    {
	arr[index].conf = unread_confs(ix)->confno();
	arr[index].unread = unread_confs(ix)->estimated_num_unread();
	index++;
    }

    assert(index == m);

    return arr;
}

long
Unread_confs::min_unread_texts_in_current()
{
    assert(current_read_order != NULL);
    return current_read_order->min_num_unread();
}

long
Unread_confs::max_unread_texts_in_current()
{
    assert(current_read_order != NULL);
    return current_read_order->max_num_unread();
}

long
Unread_confs::estimated_unread_texts_in_current()
{
    assert(current_read_order != NULL);
    return current_read_order->estimated_num_unread();
}

Read_order_iterator
Unread_confs::current_conf_begin(int wanted_texts)
{
    assert(current_read_order != NULL);
    return current_read_order->begin(wanted_texts);
}

Read_order_iterator
Unread_confs::current_conf_end()
{
    assert(current_read_order != NULL);
    return current_read_order->end();
}

Array<Text_no>
Unread_confs::query_pending_footnotes() const
{
    assert(current_read_order != NULL);
    return current_read_order->query_pending_footnotes();
}

Array<Text_no>
Unread_confs::query_pending_comments() const
{
    assert(current_read_order != NULL);
    return current_read_order->query_pending_comments();
}

void
Unread_confs::set_pending_footnotes(const Array<Text_no> &arr)
{
    assert(current_read_order != NULL);
    current_read_order->set_pending_footnotes(arr);
}

void
Unread_confs::set_pending_comments(const Array<Text_no> &arr)
{
    assert(current_read_order != NULL);
    current_read_order->set_pending_comments(arr);
}

Read_order *
Unread_confs::add_conference(Conf_no cno, bool last)
{
    if (need_unread_confs() != st_ok)
	return NULL;

    Read_order  * read_order;

    read_order = new Read_order(session, cno, reader);
    if (last)
	unread_confs.append(read_order);
    else
	unread_confs.prepend(read_order);

    return read_order;
    // Note: If there were no texts in this conference, it will be
    // moved to read_confs as soon as we start to prefetch from it.
}

Status
Unread_confs::join_conference(Conf_no cno, 
			      unsigned char priority,
			      unsigned short placement)
{
    // FIXME: There should be a way to insert the new conference in the
    // correct place. We do not know the order of the conferences. Big
    // lose.+++

    if (need_unread_confs() != st_ok)
	return st_error;

    Pix ix;
    bool present = false;
    for (ix = unread_confs.first(); ix != 0 && !present; unread_confs.next(ix))
	if (unread_confs(ix)->confno() == cno)
	    present = true;
    for (ix = read_confs.first(); ix != 0 && !present; read_confs.next(ix))
	if (read_confs(ix)->confno() == cno)
	    present = true;
    
    Status ret = session->add_member(cno, reader, priority, placement);

    if (ret != st_ok)
    {
	// FIXME: Errorcode
	return ret;
    }

    if (!present)
	add_conference(cno, true);

    return ret;
}

Kom_err
Unread_confs::disjoin_conference(Conf_no cno)			
{
    // FIXME: need_unread_confs should return error code
    if (need_unread_confs() != st_ok)
	return KOM_UNKNOWN_ERROR;

    Read_order *ro = NULL;
    Pix ix;
    for (ix = unread_confs.first(); ix != 0; unread_confs.next(ix))
	if (unread_confs(ix)->confno() == cno)
	{
	    ro = unread_confs(ix);
	    break;
	}

    Pix iy = 0;
    if (ro == NULL)
	for (iy = read_confs.first(); iy != 0; read_confs.next(iy))
	    if (read_confs(iy)->confno() == cno)
	    {
		ro = read_confs(iy);
		break;
	    }
    
    Kom_err ret = session->sub_member(cno, reader);

    if (ret == KOM_NO_ERROR)
    {
	if (current_read_order == ro)
	    current_read_order = NULL;
	if (ix != 0 && unread_confs(ix) == ro)
	    unread_confs.del(ix);
	else if (iy != 0 && read_confs(iy) == ro)
	    read_confs.del(iy);
	delete ro;
    }

    return ret;
}

Kom_err
Unread_confs::set_last_read(Conf_no cno, 
			   Local_text_no last_text_read)
{
    // FIXME: There should be a way to insert the conference in the
    // correct place (in those cases where it hasn't already been
    // fetched.  We do not know the order of the conferences.  Big
    // lose.+++

    if (need_unread_confs() != st_ok)
	return KOM_INDETERMINATE;

    Pix ix;
    Read_order *ro = NULL;
    for (ix = unread_confs.first(); ix != 0 && ro==NULL; unread_confs.next(ix))
	if (unread_confs(ix)->confno() == cno)
	    ro = unread_confs(ix);
    for (ix = read_confs.first(); ix != 0 && ro == NULL; read_confs.next(ix))
	if (read_confs(ix)->confno() == cno)
	    ro = read_confs(ix);
    
    if (ro == NULL)
	ro = add_conference(cno, true);

    if (ro == NULL)
	return KOM_INDETERMINATE;

    return ro->set_last_read(last_text_read);
}

void
Unread_confs::new_text(const Async_text_created &msg)
{
    if (need_unread_confs() != st_ok)
	return;

    Array <Conf_no> not_present;

    Text_stat t = msg.text_stat();
    for (int ir = 0; ir < t.num_recipients(); ir++)
    {
	Recipient *rcpt = t.recipient(ir);
	Conf_no confno = rcpt->recipient();
	bool present = false;

	for (Pix ix = unread_confs.first(); ix != 0 && !present;
	     unread_confs.next(ix))
	{
	    if (unread_confs(ix)->confno() == confno)
		present = true;
	}
	if (!present)
	{
	    for (Pix iy = read_confs.first(); iy != 0; read_confs.next(iy))
		if (read_confs(iy)->confno() == confno)
		{
		    unread_confs.append(read_confs(iy));
		    read_confs.del(iy);
		    present = true;
		    break;
		}
	}

	if (!present)
	{
	    Conf_no cno = rcpt->recipient();
	    not_present.push_back(cno);
	    session->prefetch_is_member(reader, cno);
	}
    }

    for (int ix = 0; ix < not_present.size(); ix++)
	if (session->is_member(reader, not_present[ix]))
	    add_conference(not_present[ix], true);
}

Status
Unread_confs::need_one_mapped()
{
    while (prefetch(&Work_quota(1,0,10)) == st_pending)
	if (Connection::drain(100) == st_error)
	    return st_error;
    return st_ok;
}

Status
Unread_confs::need_unread_confs()
{
    while (!unread_confs_has_been_fetched
	   && prefetch(&Work_quota(1,0,10)) == st_pending)
    {
	if (Connection::drain(100) == st_error)
	    return st_error;
    }

    return unread_confs_has_been_fetched ? st_ok : st_error;
}

// eof
