// Access to fileds in a user-area.
//
// $Id: user-area.cc,v 1.2 1996/08/21 09:47:28 ceder Exp $
// Copyright (C) 1996  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "create_text_q.h"
#include "set_user_area_q.h"
#include "user-area.h"
#include "text.h"
#include "LyStr.LyStr.CHMap.h"
#include "LyStr.LyStrLyStrCHMap_ptr.CHMap.h"

class User_area_internal MEMCHECK {
  public:
    User_area_internal(Person p, Text_stat_cache *, Text_mass_cache *);
    ~User_area_internal();

    // Try to fetch the userarea.  Only block if the parameter is
    // true.
    void fetch(bool block = true);

    // If we were using STL the object on the next line would have
    // been a "map<LyStr, map<LyStr, LyStr> >", which would have been
    // a lot more readable type name.  The first index is the name of
    // the area, and the second index is the key.
    LyStrLyStrLyStrCHMap_ptrCHMap parsed_map;

    // The parts of the map that don't follow the standard format set
    // out by the elisp-client are stored in this map (whose type
    // would have been "map<LyStr, LyStr>", had we used STL).
    LyStrLyStrCHMap unknown_map;

    // The last known text number of the user area.
    Text_no used_user_area;

    // Error code if the user area couldn't be fetched.
    Kom_err frozen_error_code;

    // True once the user area has been fetched (or once it is clear
    // that the user area cannot be fetched).
    bool fetch_complete;

    // True if unsaved changes has been made.
    bool unclean;

    // The person whose user-area this object handles.
    Person person;

    Text_stat_cache *tsc;
    Text_mass_cache *tmc;
};

User_area_internal::User_area_internal(Person p,
				       Text_stat_cache *tscache,
				       Text_mass_cache *tmcache)
  : parsed_map(NULL),
    unknown_map(),
    used_user_area(0),
    frozen_error_code(KOM_NO_ERROR),
    fetch_complete(false),
    unclean(false),
    person(p),
    tsc(tscache),
    tmc(tmcache)
{
    fetch(false);
}

User_area_internal::~User_area_internal()
{
}

void
User_area_internal::fetch(bool block)
{
    if (fetch_complete)
	return;

    if (!block && person.prefetch_pers_only() == st_pending)
	return;

    if (!person.pers_exists())
    {
	frozen_error_code = person.errno();
	fetch_complete = true;
	used_user_area = 0;
	return;
    }

    used_user_area = person.user_area();
    if (used_user_area == 0)
    {
	frozen_error_code = KOM_NO_USER_AREA;
	fetch_complete = true;
	return;
    }

    Text t = Text(used_user_area, tsc, tmc);

    if (!block && t.prefetch_text() == st_pending)
	return;

    LyStr_errno raw_text = t.text_mass();
    if (raw_text.errno != KOM_NO_ERROR)
    {
	frozen_error_code = raw_text.errno;
	fetch_complete = true;
	used_user_area = 0;
	return;
    }

    prot_a_LyStr userarea = raw_text.str;
    LyStr_errno block_names = userarea.stos();
    if (block_names.errno != KOM_NO_ERROR)
    {
	frozen_error_code = KOM_BAD_USER_AREA;
	fetch_complete = true;
	return;
    }

    LyStr_errno block_name;
    prot_a_LyStr block_names_a = block_names.str;
    LyStr_errno datablock_err;
    prot_a_LyStr datablock;
    LyStr_errno key;
    LyStr_errno val;
    while ((block_name = block_names_a.stos()).errno == KOM_NO_ERROR)
    {
	datablock_err  = userarea.stos();
	if (datablock_err.errno != KOM_NO_ERROR)
	{
	    frozen_error_code = datablock_err.errno;
	    fetch_complete = true;
	    return;
	}

	datablock = datablock_err.str;
	LyStrLyStrCHMap *current_map = new LyStrLyStrCHMap();
	while ((key = datablock.stos()).errno == KOM_NO_ERROR
	       && (val = datablock.stos()).errno == KOM_NO_ERROR)
	{
	    (*current_map)[key.str] = val.str;
	    if (datablock.prefix("\n"))
		datablock.remove_first();
	}
	datablock.skipspace();
	if (key.errno != KOM_BAD_HOLLERITH || datablock.strlen() != 0)
	{
	    // We failed to parse this block.
	    delete current_map;
	    unknown_map[block_name.str] = datablock_err.str;
	}
	else
	{
	    // This block was properly parsed.
	    parsed_map[block_name.str] = current_map;
	}
    }

    block_names_a.skipspace();
    userarea.skipspace();

    if (block_names_a.strlen() > 0 || userarea.strlen() > 0)
    {
	frozen_error_code = KOM_BAD_USER_AREA;
	fetch_complete = true;
	return;
    }

    fetch_complete = true;
    unclean = false;
}

User_area::User_area(Person p, Text_stat_cache *tsc, Text_mass_cache *tmc)
  : uai(new User_area_internal(p, tsc, tmc))
{
}

User_area::~User_area()
{
    for (Pix ix = uai->parsed_map.first(); ix != 0; uai->parsed_map.next(ix))
    {
	delete uai->parsed_map.contents(ix);
    }

    delete uai;
}

Kom_res<LyStr>
User_area::lookup(const LyStr &area,
		  const LyStr &key)
{
    uai->fetch();
    if (uai->frozen_error_code != KOM_NO_ERROR)
	return uai->frozen_error_code;

    Pix areapix = uai->parsed_map.seek(area);
    if (areapix == 0)
	return KOM_KEY_NOT_FOUND;
    
    LyStrLyStrCHMap *areamap = uai->parsed_map.contents(areapix);

    Pix keypix = areamap->seek(key);
    if (keypix == 0)
	return KOM_KEY_NOT_FOUND;

    return areamap->contents(keypix);
}

Kom_err
User_area::set(const LyStr &area, const LyStr &key, const LyStr &value)
{
    uai->fetch();
    if (uai->frozen_error_code == KOM_NO_USER_AREA)
	uai->frozen_error_code = KOM_NO_ERROR;
	
    if (uai->frozen_error_code != KOM_NO_ERROR)
	return uai->frozen_error_code;

    if (uai->unknown_map.contains(area))
	return KOM_BAD_USER_AREA;

    Pix areapix = uai->parsed_map.seek(area);
    LyStrLyStrCHMap *areamap;
    if (areapix == 0)
    {
	areamap = new LyStrLyStrCHMap();
	uai->parsed_map[area] = areamap;
    }
    else
    {
	areamap = uai->parsed_map.contents(areapix);
    }

    Pix keypix = areamap->seek(key);
    if (keypix == 0)
    {
	(*areamap)[key] = value;
	uai->unclean = true;
    }
    else if(areamap->contents(keypix) != value)
    {
	areamap->contents(keypix) = value;
	uai->unclean = true;
    }

    return KOM_NO_ERROR;
}

Kom_err
User_area::store_in_server()
{
    uai->fetch();
    
    if (!uai->unclean)
	return KOM_NO_ERROR;

    if (uai->frozen_error_code != KOM_NO_ERROR)
	return uai->frozen_error_code;

    prot_a_LyStr areas;
    prot_a_LyStr datablocks;

    for (Pix ix = uai->parsed_map.first(); ix != 0; uai->parsed_map.next(ix))
    {
	areas << ' ';
	areas.append_hollerith(uai->parsed_map.key(ix));
	
	prot_a_LyStr block;
	LyStrLyStrCHMap *blockmap = uai->parsed_map.contents(ix);
	for (Pix bx = blockmap->first(); bx != 0; blockmap->next(bx))
	{
	    block << ' ';
	    block.append_hollerith(blockmap->key(bx));
	    block << ' ';
	    block.append_hollerith(blockmap->contents(bx));
	    block << '\n';
	}

	datablocks.append_hollerith(block);
    }

    for (Pix iy = uai->unknown_map.first(); iy != 0; uai->unknown_map.next(iy))
    {
	areas.append_hollerith(uai->unknown_map.key(iy));
	datablocks.append_hollerith(uai->unknown_map.contents(iy));
    }
    
    prot_a_LyStr total;
    total.append_hollerith(areas);
    total << ' ' << datablocks;

    create_text_question q(uai->person.connection(), total,
			   DLList<Recipient>(),
			   DLList<Text_no>(),
			   DLList<Text_no>());
    if (q.receive() != st_ok)
	return q.error();

    set_user_area_question suq(uai->person.connection(), uai->person.conf_no(),
			       q.result());
    if (suq.receive() != st_ok)
	return suq.error();

    uai->unclean = false;

    // FIXME: update user_area and create_texts in person.
    
    uai->used_user_area = q.result();
    return KOM_NO_ERROR;
}
